import processing.core.*; 
import processing.xml.*; 

import peasy.*; 
import java.util.*; 

import java.applet.*; 
import java.awt.*; 
import java.awt.image.*; 
import java.awt.event.*; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class Boids extends PApplet {




PeasyCam cam;

/**
 * Crazy Flocking 3D Birds 
 * by Ira Greenberg. 
 * 
 * Simulates a flock of birds using a Bird class and nested
 * pushMatrix() / popMatrix() functions. 
 * Trigonometry functions handle the flapping and sinuous movement.
 * 
 * More realistic flocking behaviour, a camera and a predator added by Joren Six.
 */

// Flock array
int birdCount = 100;
Bird[]birds = new Bird[birdCount];

int predatorCount = 1;
Bird[] predators = new Bird[predatorCount];


public void setup() {
  size(screen.width,screen.height,P3D);
  noStroke();
    
  cam = new PeasyCam(this, 1000);
  cam.setMinimumDistance(50);
  cam.setMaximumDistance(50000);

  frameRate(60);
  
  for(int p = 0; p < predatorCount ; p++){
      predators[p] = new Bird(null);
  }
  
  // Initialize arrays with random values
  for (int i = 0; i < birdCount; i++){
    birds[i] = new Bird(predators);
  }
    
  for(int j = 0 ; j < birdCount;j++){  
    for(int p = 0; p < predatorCount ; p++){
      predators[p].addNeighbour(birds[j]);
    }
      
    for(int i=0 ; i<birdCount;i++){
      if(i!=j){
        birds[j].addNeighbour(birds[i]);     
      }
    }    
  }
  
}

public void draw() {
  background(0);
  lights();
  for (int i = 0; i < birdCount; i++){
    birds[i].fly();
  }
  
  for(int p = 0; p < predatorCount ; p++){
      predators[p].fly();
    }
  
}





class Bird {
  
  class ClosestNeighboursComparator implements Comparator{
    Vector sourcePosition;
    ClosestNeighboursComparator(Vector sourcePosition){
      this.sourcePosition = sourcePosition;
    }
    public int compare(Object thisBird,Object otherBird){
      float distanceToThisBird = sourcePosition.distanceTo(((Bird)thisBird).position);
      float distanceToOtherBird = sourcePosition.distanceTo(((Bird)otherBird).position);
      return new Float(distanceToThisBird).compareTo(new Float(distanceToOtherBird));
    }
  }
  
  float birdHeight = 30;
  float birdwidth = 40;
  int bodyFill = 153;
  int wingFill = 204;  
  float ang = 0;
  float flapSpeed = 0.8f;
  // Properties
  Vector position;
  Vector velocity;  
  ArrayList neighbours;
  Bird[] predators;
  
  Bird(Bird[] predators){
    this.predators = predators;
    if(predators==null){
      bodyFill= color(255,5,10);
      wingFill = color(255,5,10);
    }
      
    this.ang = ((float)random(-100,100))/100.0f;
    this.neighbours = new ArrayList();
    this.position = new Vector(random(-9000,9000),random(-9000,9000),random(-9000,0));
    this.velocity = new Vector(0.0f,0.0f,0.0f);
  }
  
  public void sortNeighbours(){
    Collections.sort(neighbours,new ClosestNeighboursComparator(position));
  }
   
   //separation
  public Vector distanceVelocity(){
    float minimumDistance = 400;
    Vector distanceVelocity = new Vector(0,0,0);
    for(int i=0;i<30;i++){
      Bird neighbour = (Bird) neighbours.get(i);
      if(position.distanceTo(neighbour.position) < minimumDistance)
          distanceVelocity = distanceVelocity.subtract(neighbour.position.subtract(this.position));
    }
    distanceVelocity = distanceVelocity.devide(30);
    return distanceVelocity.normalize();
  }
  
  public Vector attackClosestVelocity(){
    float minimumDistance = 800;
    Vector distanceVelocity = new Vector(0,0,0);
    Bird neighbour = (Bird) neighbours.get(0);
    if(position.distanceTo(neighbour.position) < minimumDistance)
          distanceVelocity = distanceVelocity.add(neighbour.position.add(this.position));
    
    return distanceVelocity.normalize();
  }
  
  
  //cohesion
  public Vector centerVelocity(){
    Vector centerVelocity = new Vector(0,0,0);
    for(int i=0;i<neighbours.size();i++){      
      Bird neighbour = (Bird) neighbours.get(i % 5 == 0 ? (int) random(0,10) : i);
      centerVelocity = centerVelocity.add(neighbour.position);
    }
    centerVelocity = centerVelocity.devide(100);
    return centerVelocity.subtract(position).normalize();
  }
  
  public Vector centerOfBoxVelocity(){
    Vector centerVelocity = new Vector(0,0,0);
    return centerVelocity.subtract(position).normalize();
  }
  
  public Vector avoidPredators(){
    Vector avoidPredatorVelocity = new Vector(0.0f,0.0f,0.0f);
    for(int p=0;p < predators.length ; p++){
      Bird predator = predators[p];
      Vector predatorPlace = predator.position;
      if(predator.position.distanceTo(position) < 3000)
        return position.subtract(predator.position).normalize();
    }
    return avoidPredatorVelocity;
  }
  
  //alignment
  public Vector similarVelocity(){
    Vector similarVelocity = new Vector(0.0f,0.0f,0.0f);
    for(int i=0;i<5;i++){
      Bird neighbour = (Bird) neighbours.get(i);
      similarVelocity = similarVelocity.add(neighbour.velocity);
    }
    similarVelocity = similarVelocity.devide(5);
    return similarVelocity.subtract(velocity).normalize();
  }
  
  public Vector limitVelocity(Vector velocity){
    float maxSpeed = predators == null ? 440 : 400;
    float speed = velocity.size();
    if(speed > maxSpeed)
      return velocity.devide(speed).multiply(maxSpeed);
    return velocity;
  }
  
  public Vector boundingBoxVelocity(){
    int maxX = width * 30;
    int maxY = height * 30;
    int maxZ = height * 30;
    int force = 70;
    
    float velocity_x = 0.0f;    
    if(position.x < -1 * maxX)
      velocity_x = force;
    else if(position.x > maxX)
      velocity_x =  force * -1;
      
    float velocity_y = 0.0f;
    if(position.y < -1 * maxY)
      velocity_y = force ;
    else if(position.y > maxY)
      velocity_y = force * -1;      
      
    float velocity_z = 0.0f;
    if(position.z < -1 * maxZ)
      velocity_z = force;
    else if(position.z > maxZ)
      velocity_z =  -1 * force;
    
    return new Vector(velocity_x,velocity_y,velocity_z).normalize();
  }
    
  public void addNeighbour(Bird otherBird){
    this.neighbours.add(otherBird);
  }


  public void fly() {
    pushMatrix();
    
    try{
    //Thread.sleep(20);
    }catch(Exception e){
    }
    
    sortNeighbours();
    /*
     System.out.println("position:\t\t" + position);  
    System.out.println("bounding:\t\t" +boundingBoxVelocity());
    System.out.println("center v:\t\t" +centerVelocity().devide(1000));
    System.out.println("distancev:\t\t" + distanceVelocity());
    */
    if(predators == null){
      velocity = limitVelocity(
      attackClosestVelocity().multiply(55)
      .add(centerVelocity().multiply(40))
      );
    }else{
      velocity = limitVelocity(
     boundingBoxVelocity().multiply(56)
    .add(centerVelocity().multiply(25))
    .add(distanceVelocity().multiply(55))
    .add(similarVelocity().multiply(45))
    .add(avoidPredators().multiply(48))
    .add(centerOfBoxVelocity().multiply(25))
    );
/*
    System.out.println("predatorp:\t\t" + predator.position);    
    System.out.println("predatorv:\t\t" + avoidPredator());
    System.out.println("similar v:\t\t" + similarVelocity());    
    System.out.println("total v:\t\t" + velocity);
 */ 
    
    }
    
    position = position.add(velocity);
    //System.out.println("new position:\t" + position + "\n");
    
    translate(position.x,position.y,position.z-5000);
    
    

    rotateX(PI/8);
    rotateY(PI/2);//
    rotateZ(PI/2);//position.getZAngle()

    if(predators == null){
     // System.out.println("new position:\t" + position + "\n");
    // Body
    fill(bodyFill);
    box(20,100,20);
    
    //System.out.println(PI/4);
    //System.out.println(new Vector(1.0,1.0,0.0).getXAngle());
    //System.out.println(new Vector(1.0,1.0,0.0).getYAngle());
    //System.out.println(new Vector(1.0,1.0,0.0).getZAngle());

    // Left wing
    fill(wingFill);
    pushMatrix();
    rotateY(sin(radians(ang)) * 20);
    rect(-75, -50, 75, 100);
    popMatrix();

    // Right wing
    pushMatrix();
    rotateY(sin(radians(ang)) * -20);
    rect(-75, -50, 75, 100);
    popMatrix();
    
    }else{
      
    // Body
    fill(bodyFill);
    box(40,200,40);

    // Left wing
    fill(wingFill);
    pushMatrix();
    rotateY(sin(radians(ang)) * 20);
    rect(-150, -100, 150, 200);
    popMatrix();

    // Right wing
    pushMatrix();
    rotateY(sin(radians(ang)) * -20);
    rect(-150, -100, 150, 200);
    popMatrix();  
    }

    // Wing flap
    double factor = 1;
    if (ang > 2) {
      factor*=-1;
    } 
    if (ang < -2) {
      factor*=-1;
    }

    ang += (float)velocity.size() / 50.0f  ;
    
    popMatrix();
  }
}





class Vector{
  
  final float x,y,z;
  
  Vector(float x,float y,float z){
    this.x=x;
    this.y=y;
    this.z=z;  
  }
  
  public Vector multiply(float scalar){
      return new Vector(x * scalar, y*scalar,z*scalar);
  }
  
  public Vector devide(float scalar){
    return new Vector(x / scalar, y/scalar,z/scalar);
  }
  
  public Vector add(Vector other){
    return new Vector(x + other.x,y + other.y,z + other.z);
  }
  
  public Vector subtract(Vector other){
    return new Vector(x - other.x,y-other.y,z-other.z);
  }
  
  public float distanceTo(Vector other){
    return (float) Math.pow((x - other.x) * (x - other.x) + (y - other.y) * (y - other.y) + (z - other.z) * (z - other.z),0.5f);
  }
  
  public Vector normalize(){
    if(size()==0.0f)
      return this;
    else
      return devide(size());
  }
  
  public float size(){
    return distanceTo(new Vector(0,0,0));
  }
  
  public String toString(){
    return x + "\t,\t" + y + "\t,\t" + z;
  }   
  
  public float getXAngle(){
      return acos(x/size());
  }
  
  public float getYAngle(){
      return acos(y/size());
  }
  
  public float getZAngle(){
      return acos(z/size());
  }
}

  static public void main(String args[]) {
    PApplet.main(new String[] { "--present", "--bgcolor=#666666", "--stop-color=#cccccc", "Boids" });
  }
}
